import logging
from typing import Union

import faiss
import numpy as np
import torch
from sklearn.neighbors import NearestNeighbors
from tqdm import tqdm

logger = logging.getLogger(__name__)


class AutoModelForMatch(object):
    def __init__(self, method: str = "cosine") -> None:
        super().__init__()
        self.method = method

    def similarity_search(
        self,
        query_embed: torch.Tensor,
        passage_embed: torch.Tensor,
        top_k: int = 1,
        batch_size: int = 0,
        convert_to_numpy: bool = True,
        **kwargs,
    ):
        if self.method == "knn":
            neighbors_model = NearestNeighbors(n_neighbors=top_k, metric="cosine", n_jobs=-1)
            neighbors_model.fit(passage_embed)
            dists, indices = neighbors_model.kneighbors(query_embed)
            return dists, indices

        elif self.method == "cosine":
            dists, indices = cosine_similarity_search(
                query_embed, passage_embed, top_k=top_k, batch_size=batch_size, convert_to_numpy=convert_to_numpy
            )
            return dists, indices

        else:
            raise ValueError(f"Only cosine and knn method are supported by similarity_search, while get {self.method}")

    def faiss_search(
        self,
        query_embed: torch.Tensor,
        index_path: str = "/faiss.index",
        top_k: int = 1,
        batch_size: int = 256,
        max_length: int = 512,
    ):
        faiss_index = faiss.read_index(index_path)
        dists, indices = faiss_search(
            query_embeddings=query_embed,
            faiss_index=faiss_index,
            top_k=top_k,
            batch_size=batch_size,
        )
        return dists, indices

    def get_match_df(self):
        rerank_data = dict({'query': [], 'passage': [], 'labels': []})
        return rerank_data


def cosine_similarity_search(
    query_embed: torch.Tensor,
    passage_embed: torch.Tensor,
    top_k: int = 1,
    batch_size: int = 512,
    penalty: bool = True,
    temperature: float = 0,
    convert_to_numpy: bool = True,
):
    if len(query_embed.size()) == 1:
        query_embed = query_embed.view(1, -1)
    assert query_embed.size()[1] == passage_embed.size()[1], (
        f"The embed Shape of query_embed and passage_embed should be same, "
        f"while received query {query_embed.size()} and passage {passage_embed.size()}"
    )
    chunk = batch_size if batch_size > 0 else len(query_embed)
    embeddings_chunks = query_embed.split(chunk)

    dists = []
    indices = []
    for idx in range(len(embeddings_chunks)):
        cos_sim_chunk = torch.matmul(embeddings_chunks[idx], passage_embed.transpose(0, 1))
        cos_sim_chunk = torch.nan_to_num(cos_sim_chunk, nan=0.0)
        # if penalty:
        #     pen = ((contents["old_source_count"].values==0) & (contents["old_nonsource_count"].values==1))
        #     cos_sim_chunk = cos_sim_chunk.clamp(0,1) ** torch.Tensor((1 + pen*0.5).reshape(1,-1)).to(DEVICE)
        #     cos_sim_chunk = cos_sim_chunk.clamp(0,1) ** torch.Tensor((1-(contents["old_count"].values==0)*0.1).
        #     reshape(1,-1)).to(DEVICE)
        if temperature:
            cos_sim_chunk = cos_sim_chunk / temperature
        top_k = min(top_k, cos_sim_chunk.size(1))
        dists_chunk, indices_chunk = torch.topk(cos_sim_chunk, k=top_k, dim=1)
        dists.append(dists_chunk[:, :].detach().cpu())
        indices.append(indices_chunk[:, :].detach().cpu())
    dists = torch.cat(dists)
    indices = torch.cat(indices)
    if convert_to_numpy:
        dists = dists.numpy()
        indices = indices.numpy()
    return dists, indices


def faiss_search(
    query_embeddings,
    faiss_index: faiss.Index,
    top_k: int = 100,
    batch_size: int = 256,
    max_length: int = 512,
):
    """
    1. Encode queries into dense embeddings;
    2. Search through faiss index
    """
    # query_embeddings = model.encode_queries(
    #     queries["query"], batch_size=batch_size, max_length=max_length
    # )
    query_size = len(query_embeddings)

    all_scores = []
    all_indices = []

    for i in tqdm(range(0, query_size, batch_size), desc="Searching"):
        j = min(i + batch_size, query_size)
        query_embedding = query_embeddings[i:j]
        score, index = faiss_index.search(query_embedding.astype(np.float32), k=top_k)
        all_scores.append(score)
        all_indices.append(index)

    all_scores = np.concatenate(all_scores, axis=0)
    all_indices = np.concatenate(all_indices, axis=0)
    return all_scores, all_indices


class FaissIndex:
    def __init__(self, device) -> None:
        if isinstance(device, torch.device):
            if device.index is None:
                device = "cpu"
            else:
                device = device.index
        self.device = device

    def build(self, encoded_corpus, index_factory, metric):
        if metric == "l2":
            metric = faiss.METRIC_L2
        elif metric in ["ip", "cos"]:
            metric = faiss.METRIC_INNER_PRODUCT
        else:
            raise NotImplementedError(f"Metric {metric} not implemented!")

        index = faiss.index_factory(encoded_corpus.shape[1], index_factory, metric)

        if self.device != "cpu":
            co = faiss.GpuClonerOptions()
            co.useFloat16 = True
            # logger.info("using fp16 on GPU...")
            index = faiss.index_cpu_to_gpu(faiss.StandardGpuResources(), self.device, index, co)

        logging.info("training index...")
        index.train(encoded_corpus)
        logging.info("adding embeddings...")
        index.add(encoded_corpus)
        self.index = index
        return index

    def load(self, index_path):
        logging.info(f"loading index from {index_path}...")
        index = faiss.read_index(index_path)
        if self.device != "cpu":
            co = faiss.GpuClonerOptions()
            co.useFloat16 = True
            index = faiss.index_cpu_to_gpu(faiss.StandardGpuResources(), self.device, index, co)
        self.index = index
        return index

    def save(self, index_path):
        logging.info(f"saving index at {index_path}...")
        if isinstance(self.index, faiss.GpuIndex):
            index = faiss.index_gpu_to_cpu(self.index)
        else:
            index = self.index
        faiss.write_index(index, index_path)

    def search(self, query, hits):
        return self.index.search(query, k=hits)



"""
trainable temperature parameter: Text and Code Embeddings by Contrastive Pre-Training
"""

import logging

import torch
from torch import nn

logger = logging.getLogger(__name__)


class CosineSimilarity(nn.Module):
    def __init__(self, temperature: float = 0.0):
        super().__init__()
        self.temperature = temperature

    def forward(self, query_embeddings: torch.Tensor, passage_embeddings: torch.Tensor):
        sim_pos_vector = torch.cosine_similarity(query_embeddings, passage_embeddings, dim=-1)
        sim_pos_vector = sim_pos_vector / self.temperature
        sim_neg_matrix = torch.cosine_similarity(
            query_embeddings.unsqueeze(1),
            passage_embeddings.unsqueeze(0),
            dim=-1,
        )
        sim_neg_matrix = sim_neg_matrix / self.temperature
        sim_diff_matrix = sim_pos_vector.unsqueeze(1) - sim_neg_matrix
        loss = -torch.log(torch.sigmoid(sim_diff_matrix)).mean()
        return loss

    def get_temperature(self):
        if not self.dynamic_temperature:
            return self.temperature
        return torch.clamp(self.temperature, min=1e-3)


class Similarity(nn.Module):
    """
    Dot product or cosine similarity
    """

    def __init__(self, temp=0.05):
        super().__init__()
        self.temp = temp
        self.cos = nn.CosineSimilarity(dim=-1)

    def forward(self, x, y):
        return self.cos(x, y) / self.temp



from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Sequence

from langchain.callbacks.manager import Callbacks
from langchain.retrievers.document_compressors.base import BaseDocumentCompressor
from langchain.text_splitter import (
    CharacterTextSplitter,
    MarkdownHeaderTextSplitter,
    MarkdownTextSplitter,
)
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import Extra, root_validator

from src.retrievals.models.embedding_auto import AutoModelForEmbedding
from src.retrievals.models.rerank import RerankModel


class LangchainEmbedding(AutoModelForEmbedding, Embeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class LangchainReranker(BaseDocumentCompressor):
    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def __init__(
        self,
        model_name_or_path: str,
        top_n: int = 3,
        device: str = "cuda",
        max_length: int = 1024,
        batch_size: int = 32,
        show_progress_bar: bool = None,
        num_workers: int = 0,
    ):
        self._model = RerankModel(model_name_or_path=model_name_or_path, max_length=max_length, device=device)
        super().__init__(
            model=model_name_or_path,
            top_n=top_n,
            device=device,
            max_length=max_length,
            batch_size=batch_size,
            num_workers=num_workers,
        )

    def compress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Optional[Callbacks] = None,
    ) -> Sequence[Document]:
        """
        Compress documents using `BCEmbedding RerankerModel API`.

        Args:
            documents: A sequence of documents to compress.
            query: The query to use for compressing the documents.
            callbacks: Callbacks to run during the compression process.

        Returns:
            A sequence of compressed documents.
        """
        if len(documents) == 0:  # to avoid empty api call
            return []
        doc_list = list(documents)

        passages = []
        valid_doc_list = []
        invalid_doc_list = []
        for d in doc_list:
            passage = d.page_content
            if isinstance(passage, str) and len(passage) > 0:
                passages.append(passage.replace('\n', ' '))
                valid_doc_list.append(d)
            else:
                invalid_doc_list.append(d)

        rerank_result = self._model.rerank(query, passages)
        final_results = []
        for score, doc_id in zip(rerank_result['rerank_scores'], rerank_result['rerank_ids']):
            doc = valid_doc_list[doc_id]
            doc.metadata["relevance_score"] = score
            final_results.append(doc)
        for doc in invalid_doc_list:
            doc.metadata["relevance_score"] = 0
            final_results.append(doc)

        final_results = final_results[: self.top_n]
        return final_results


class RagFeature(object):
    def __init__(self, config_path: str = 'config.ini'):
        pass



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
from transformers import BatchEncoding, DataCollatorWithPadding, PreTrainedTokenizer


class PairCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer,
        max_length: Optional[int] = None,
        query_max_length: Optional[int] = None,
        passage_max_length: Optional[int] = None,
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_max_length: int
        self.passage_max_length: int
        if query_max_length:
            self.query_max_length = query_max_length
        elif max_length:
            self.query_max_length = max_length
            self.passage_max_length = max_length
        else:
            self.query_max_length = tokenizer.model_max_length
            self.passage_max_length = tokenizer.model_max_length

        if passage_max_length:
            self.passage_max_length = passage_max_length

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        assert (
            'query' in features[0] and 'pos' in features[0]
        ), "PairCollator should have 'query' and 'pos' keys in features dict"

        query_texts = [feature["query"] for feature in features]
        pos_texts = [feature["pos"] for feature in features]

        query_inputs = self.tokenizer(
            query_texts,
            padding=True,
            max_length=self.query_max_length,
            truncation=True,
            return_tensors="pt",
        )
        pos_inputs = self.tokenizer(
            pos_texts,
            padding=True,
            max_length=self.passage_max_length,
            truncation=True,
            return_tensors="pt",
        )

        return {"query": query_inputs, "pos": pos_inputs}


class TripletCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer,
        max_length: Optional[int] = None,
        query_max_length: Optional[int] = None,
        passage_max_length: Optional[int] = None,
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_max_length: int
        self.passage_max_length: int
        if query_max_length:
            self.query_max_length = query_max_length
        elif max_length:
            self.query_max_length = max_length
            self.passage_max_length = max_length
        else:
            self.query_max_length = tokenizer.model_max_length
            self.passage_max_length = tokenizer.model_max_length

        if passage_max_length:
            self.passage_max_length = passage_max_length

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        assert (
            'query' in features[0] and 'pos' in features[0] and 'neg' in features[0]
        ), "TripletCollator should have 'query', 'pos' and 'neg' keys in features dict"

        query_texts = [feature["query"] for feature in features]
        pos_texts = [feature["pos"] for feature in features]
        neg_texts = [feature["neg"] for feature in features]

        # if isinstance(query[0], list):
        #     query = sum(query, [])
        # if isinstance(passage[0], list):
        #     passage = sum(passage, [])

        query_inputs = self.tokenizer(
            query_texts,
            padding=True,
            max_length=self.query_max_length,
            truncation=True,
            return_tensors="pt",
        )
        pos_inputs = self.tokenizer(
            pos_texts,
            padding=True,
            max_length=self.passage_max_length,
            truncation=True,
            return_tensors="pt",
        )  # ["input_ids"]
        neg_inputs = self.tokenizer(
            neg_texts,
            padding=True,
            max_length=self.passage_max_length,
            truncation=True,
            return_tensors="pt",
        )  # ["input_ids"]

        return {
            "query": query_inputs,
            "pos": pos_inputs,
            "neg": neg_inputs,
        }


class RerankCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer,
        max_length: Optional[int] = None,
        query_max_length: Optional[int] = None,
        passage_max_length: Optional[int] = None,
    ):
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_max_length: int
        self.passage_max_length: int
        if query_max_length:
            self.query_max_length = query_max_length
        elif max_length:
            self.query_max_length = max_length
            self.passage_max_length = max_length
        else:
            self.query_max_length = tokenizer.model_max_length
            self.passage_max_length = tokenizer.model_max_length

        if passage_max_length:
            self.passage_max_length = passage_max_length

    def __call__(self, features: List[Dict[str, Any]]) -> BatchEncoding:
        assert (
            'query' in features and 'passage' in features
        ), "RerankCollator should have 'query' and 'passage' keys in features dict, and 'labels' during training"

        query_texts = [feature["query"] for feature in features]
        passage_texts = [feature['passage'] for feature in features]

        labels = None
        if 'labels' in features[0].keys():
            labels = [feature['labels'] for feature in features]

        batch = self.tokenizer(
            text=query_texts, text_pair=passage_texts, truncation=True, max_length=self.max_length, return_tensors="pt"
        )

        # for key in ['input_ids', 'attention_mask']:
        #     batch[key] = torch.tensor(batch[key], dtype=torch.int64)

        if labels is not None:
            batch['labels'] = torch.tensor(batch['labels'], dtype=torch.float32)
        return batch



import logging
import math
import os
import time
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch.nn as nn
from transformers import Trainer
from transformers.modeling_utils import PreTrainedModel
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from transformers.trainer_callback import TrainerCallback
from transformers.trainer_utils import EvalPrediction

from src.retrievals.losses.triplet import TripletLoss

logger = logging.getLogger(__name__)


class RetrievalTrainer(Trainer):
    def __init__(self, loss_fn, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = loss_fn

    def compute_loss(self, model, inputs, return_outputs=False, **kwargs):
        query = inputs["query"]
        pos = inputs["pos"]
        query_embeddings = model(query)
        pos_embeddings = model(pos)
        if 'neg' in inputs:
            neg = inputs["neg"]
            neg_embeddings = model(neg)
            loss = TripletLoss()(query_embeddings, pos_embeddings, neg_embeddings)
        else:
            loss = self.loss_fn(query_embeddings, pos_embeddings)
        if not return_outputs:
            return loss
        outputs = dict()
        outputs['query'] = query_embeddings
        outputs['pos'] = pos_embeddings
        if 'neg' in inputs:
            outputs['neg'] = neg_embeddings
        return (loss, outputs)

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)

        self.model.model.save_pretrained(output_dir)
        self.model.tokenizer.save_pretrained(output_dir)


class RerankTrainer(Trainer):
    def __init__(self, loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        if not loss_fn:
            loss_fn = nn.BCEWithLogitsLoss()
        self.loss_fn = loss_fn

    def compute_loss(self, model, inputs, **kwargs):
        return

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        pass



import logging
from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Tuple, Union

import faiss
import numpy as np
import torch
import torch.nn as nn
from numpy import ndarray
from torch.utils.data import DataLoader, Dataset
from tqdm.autonotebook import trange
from transformers import (
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    BatchEncoding,
    GenerationConfig,
)

from src.retrievals.models.pooling import AutoPooling

logger = logging.getLogger(__name__)


def get_device_name() -> Literal["mps", "cuda", "cpu"]:
    """
    Returns the name of the device where this module is running on.
    It's a simple implementation that doesn't cover cases when more powerful GPUs are available and
    not a primary device ('cuda:0') or MPS device is available, but not configured properly:
    https://pytorch.org/docs/master/notes/mps.html

    :return: Device name, like 'cuda' or 'cpu'
    """
    if torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    else:
        return "cpu"


def batch_to_device(batch: Dict, target_device: str) -> Dict[str, torch.Tensor]:
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], torch.Tensor):
            batch[key] = batch[key].to(target_device)
        else:
            print(batch[key])
            batch[key] = torch.tensor(batch[key], dtype=torch.long).to(target_device)
    return batch


class AutoModelForEmbedding(nn.Module):
    """
    Loads or creates a Embedding model that can be used to map sentences / text.

    :param model_name_or_path: If it is a filepath on disc, it loads the model from that path. If it is not a path,
        it first tries to download a pre-trained SentenceTransformer model. If that fails, tries to construct a model
        from the Hugging Face Hub with that name.
    """

    encode_kwargs: Dict[str, Any] = dict()
    show_progress: bool = False

    def __init__(
        self,
        model_name_or_path: str,
        pretrained: bool = True,
        config_path: Optional[str] = None,
        pooling_method: str = "cls",
        normalize_embeddings: bool = False,
        max_length: Optional[int] = None,
        loss_fn: Optional[Callable] = None,
        query_instruction: Optional[str] = None,
        passage_instruction: Optional[str] = None,
        generation_args: Dict = None,
        use_fp16: bool = False,
        use_lora: bool = False,
        lora_config=None,
        device: Optional[str] = None,
        trust_remote_code: bool = False,
    ):
        super().__init__()

        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, return_tensors=False, trust_remote_code=trust_remote_code
        )

        if pretrained:
            self.model = AutoModel.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code)
        else:
            if config_path is None:
                self.config = AutoConfig.from_pretrained(
                    model_name_or_path, output_hidden_states=True, trust_remote_code=trust_remote_code
                )
                self.model = AutoModel.from_config(self.config)
        self.loss_fn = loss_fn

        if max_length is None:
            if (
                hasattr(self.model, "config")
                and hasattr(self.model.config, "max_position_embeddings")
                and hasattr(self.tokenizer, "model_max_length")
            ):
                max_length = min(self.model.config.max_position_embeddings, self.tokenizer.model_max_length)

        self.max_length = max_length

        if use_fp16:
            self.model.half()
        if use_lora:
            # peft config and wrapping
            from peft import LoraConfig, TaskType, get_peft_model

            if not lora_config:
                raise ValueError("If use_lora is true, please provide a valid lora_config")
            self.model = get_peft_model(self.model, lora_config)
            self.model.print_trainable_parameters()

        if device is None:
            self.device = get_device_name()
        else:
            self.device = device

        self.normalize_embeddings = normalize_embeddings
        self.pooling = AutoPooling(pooling_method)

        # self.fc = nn.Linear(768, 1)
        # self._init_weights(self.fc)

        self.query_instruction = query_instruction
        self.passage_instruction = passage_instruction
        if generation_args is not None:
            generation_config = self.model.generation_config.to_dict()
            generation_config.update(generation_args)
            generation_config.update({"pad_token_id": self.tokenizer.pad_token_id})
            self.model.generation_config = GenerationConfig(**generation_config)

    def _init_weights(self, module: nn.Module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
        self,
        inputs,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = False,
    ):
        if isinstance(inputs, (dict, BatchEncoding)):
            embeddings = self.forward_from_loader(inputs)
        elif isinstance(inputs, str) or (isinstance(inputs, list) and isinstance(inputs[0], str)):
            embeddings = self.forward_from_text(inputs)
        else:
            raise ValueError

        if labels is None or self.loss_fn is None:
            if return_dict:
                return {"sentence_embedding": embeddings}
            return embeddings
        else:
            outputs = dict()
            loss_output = self.loss_fn(embeddings, labels)
            outputs["loss"] = loss_output["loss"]
            outputs["sentence_embedding"] = loss_output["sentence_embedding"]
            return outputs

    def forward_from_loader(self, inputs):
        model_output = self.model(inputs['input_ids'], inputs['attention_mask'])
        if self.pooling is not None:
            embeddings = self.pooling(model_output[0], inputs["attention_mask"])

            if self.normalize_embeddings:
                embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)

            return embeddings

    def forward_from_text(self, texts):
        batch_dict = self.tokenizer(
            texts,
            max_length=self.max_length,
            return_attention_mask=False,
            padding=False,
            truncation=True,
        )
        batch_dict["input_ids"] = [input_ids + [self.tokenizer.eos_token_id] for input_ids in batch_dict["input_ids"]]
        batch_dict = self.tokenizer.pad(batch_dict, padding=True, return_attention_mask=True, return_tensors="pt")
        batch_dict.pop("token_type_ids")
        return self.forward_from_loader(batch_dict)

    def encode(
        self,
        inputs,
        batch_size: int = 32,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
    ):
        if isinstance(inputs, (BatchEncoding, Dict)):
            return self.encode_from_loader(
                loader=inputs,
                batch_size=batch_size,
                show_progress_bar=show_progress_bar,
                output_value=output_value,
                convert_to_numpy=convert_to_numpy,
                convert_to_tensor=convert_to_tensor,
                device=device,
                normalize_embeddings=normalize_embeddings,
            )
        elif isinstance(inputs, (str, List, Tuple)):
            return self.encode_from_text(
                sentences=inputs,
                batch_size=batch_size,
                show_progress_bar=show_progress_bar,
                output_value=output_value,
                convert_to_numpy=convert_to_numpy,
                convert_to_tensor=convert_to_tensor,
                device=device,
                normalize_embeddings=normalize_embeddings,
            )
        else:
            raise ValueError

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a HuggingFace transformer model."""
        embeddings = self.encode(texts, show_progress_bar=self.show_progress, **self.encode_kwargs)
        return embeddings.tolist()

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a HuggingFace transformer model."""
        return self.embed_documents([text])[0]

    def encode_from_loader(
        self,
        loader,
        batch_size: int = 32,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
    ) -> Union[List[torch.Tensor], ndarray, torch.Tensor]:
        self.model.eval()
        self.model.to(device or self.device)
        all_embeddings = []

        with torch.no_grad():
            for idx, inputs in enumerate(loader):
                for k, v in inputs.items():
                    inputs[k] = v.to(self.device)
                embed = self.forward_from_loader(inputs)
                embed = embed.detach().cpu()
                if convert_to_numpy:
                    embed = embed.numpy()
                all_embeddings.append(embed.numpy())
        if convert_to_numpy:
            all_embeddings = np.concatenate(all_embeddings)
        else:
            all_embeddings = torch.concat(all_embeddings)
        return all_embeddings

    def encode_from_text(
        self,
        sentences: Union[str, List[str]],
        batch_size: int = 32,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
    ) -> Union[List[torch.Tensor], ndarray, torch.Tensor]:
        """
        Computes sentence embeddings from sentence-transformers library.

        :param sentences: the sentences to embed.
        :param batch_size: the batch size used for the computation.
        :param show_progress_bar: Whether to output a progress bar when encode sentences.
        :param output_value: The type of embeddings to return: "sentence_embedding" to get sentence embeddings,
            "token_embeddings" to get wordpiece token embeddings, and `None`, to get all output values. Defaults
            to "sentence_embedding".
        :param convert_to_numpy: Whether the output should be a list of numpy vectors. If False, it is a list of tensors
        :param convert_to_tensor: Whether the output should be one large tensor. Overwrites `convert_to_numpy`.
        :param device: Which `torch.device` to use for the computation.
        :param normalize_embeddings: Whether to normalize returned vectors to have length 1. In that case,
            the faster dot-product (util.dot_score) instead of cosine similarity can be used.

        :return: By default, a list of tensors is returned. If convert_to_tensor, a stacked tensor is returned.
            If convert_to_numpy, a numpy matrix is returned.
        """
        self.model.eval()
        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG
            )

        if convert_to_tensor:
            convert_to_numpy = False

        if output_value != "sentence_embedding":
            convert_to_tensor = False
            convert_to_numpy = False

        input_was_string = False
        if isinstance(sentences, str) or not hasattr(
            sentences, "__len__"
        ):  # Cast an individual sentence to a list with length 1
            sentences = [sentences]
            input_was_string = True

        if device is None:
            device = self.device

        self.to(device)

        all_embeddings = []
        length_sorted_idx = np.argsort([-self._text_length(sentence) for sentence in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index : start_index + batch_size]
            features = self.tokenizer(
                sentences_batch,
                padding=True,
                truncation="longest_first",
                return_tensors="pt",
                max_length=self.max_length,
            )
            features = batch_to_device(features, device)

            with torch.no_grad():
                out_features = self.forward(features, return_dict=True)

                if output_value == "token_embeddings":
                    embeddings = []
                    for token_emb, attention in zip(out_features[output_value], out_features["attention_mask"]):
                        last_mask_id = len(attention) - 1
                        while last_mask_id > 0 and attention[last_mask_id].item() == 0:
                            last_mask_id -= 1

                        embeddings.append(token_emb[0 : last_mask_id + 1])
                elif output_value is None:  # Return all outputs
                    embeddings = []
                    for sent_idx in range(len(out_features["sentence_embedding"])):
                        row = {name: out_features[name][sent_idx] for name in out_features}
                        embeddings.append(row)
                else:  # Sentence embeddings
                    embeddings = out_features[output_value]
                    embeddings = embeddings.detach()
                    if normalize_embeddings:
                        embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)

                    # fixes for #522 and #487 to avoid oom problems on gpu with large datasets
                    if convert_to_numpy:
                        embeddings = embeddings.cpu()

                all_embeddings.extend(embeddings)

        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]

        if convert_to_tensor:
            if len(all_embeddings):
                all_embeddings = torch.stack(all_embeddings)
            else:
                all_embeddings = torch.Tensor()
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_was_string:
            all_embeddings = all_embeddings[0]

        return all_embeddings

    def build_index(self, inputs: BatchEncoding, batch_size: int = 64, use_gpu: bool = True):
        embeddings = self.encode(inputs, batch_size=batch_size)
        embeddings = np.asarray(embeddings, dtype=np.float32)
        index = faiss.IndexFlatL2(len(embeddings[0]))
        if use_gpu:
            co = faiss.GpuMultipleClonerOptions()
            co.shard = True
            co.useFloat16 = True
            index = faiss.index_cpu_to_all_gpus(index, co=co)
        index.add(embeddings)
        return index

    def add_to_index(self):
        return

    def search(self):
        return

    def similarity(self, queries: Union[str, List[str]], keys: Union[str, List[str], ndarray]):
        return

    def save(self):
        pass

    @classmethod
    def from_pretrained(cls, model_name_or_path: str, **kwargs):
        embedder = AutoModel.from_pretrained(model_name_or_path)
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        return cls(embedder, tokenizer, **kwargs)

    def save_pretrained(self, output_path: str):
        self.tokenizer.save_pretrained(output_path)
        self.model.config.save_pretrained(output_path)

    def _text_length(self, text: Union[List[int], List[List[int]]]):
        """
        Help function to get the length for the input text. Text can be either
        a list of ints (which means a single text as input), or a tuple of list of ints
        (representing several text inputs to the model).
        """

        if isinstance(text, dict):  # {key: value} case
            return len(next(iter(text.values())))
        elif not hasattr(text, "__len__"):  # Object has no len() method
            return 1
        elif len(text) == 0 or isinstance(text[0], int):  # Empty string or list of ints
            return len(text)
        else:
            return sum([len(t) for t in text])  # Sum of length of individual strings


class PairwiseModel(AutoModelForEmbedding):
    """Pairwise Model wrapper
    - bi_encoder
    - cross_encoder
    - poly_encoder
    support: query + pos pair, or query + pos + neg triplet
    """

    def __init__(
        self,
        model_name_or_path: str,
        pooling_method: str = "cls",
        normalize_embeddings: bool = False,
        query_instruction: Optional[str] = None,
        use_fp16: bool = False,
        cross_encoder: bool = False,
        poly_encoder: bool = False,
        temperature: float = 0,
        dynamic_temperature: bool = False,
        loss_fn: Union[nn.Module, Callable] = None,
    ) -> None:
        super().__init__(
            model_name_or_path=model_name_or_path,
            pooling_method=pooling_method,
            normalize_embeddings=normalize_embeddings,
            query_instruction=query_instruction,
            use_fp16=use_fp16,
            loss_fn=None,
        )
        if loss_fn is not None:
            logger.warning("loss_fn in Pairwise embed model, which will be ignored")

        self.cross_encoder = cross_encoder
        self.temperature = temperature
        self.dynamic_temperature = dynamic_temperature

    def forward(
        self,
        inputs: List[torch.Tensor],
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        input1 = inputs[0]
        input2 = inputs[1]
        if len(inputs) == 3:
            input3 = inputs[2]

        if self.cross_encoder:
            ids1, mask1 = input1
            ids2, mask2 = input2
            ids = torch.cat([ids1, ids2], dim=0)
            mask = torch.cat([mask1, mask2], dim=0)

            transformer_out = super().forward({"input_ids": ids, "attention_mask": mask})
            pooled_output = self.pooling(transformer_out[0], mask)
            pooled_output1 = pooled_output[: len(ids1), :]
            pooled_output2 = pooled_output[len(ids1) :, :]
            return pooled_output1, pooled_output2
        else:
            # bi-encoder, pooling in each
            pooled_output1 = super().forward(input1)
            pooled_output2 = super().forward(input2)
            if len(inputs) == 3:
                pooled_output3 = super().forward(input3)
                return pooled_output1, pooled_output2, pooled_output3
            return pooled_output1, pooled_output2


def unsorted_segment_mean(data: torch.Tensor, segment_ids: torch.Tensor, num_segments: int) -> torch.Tensor:
    result_shape = (num_segments, data.size(1))
    segment_ids = segment_ids.unsqueeze(-1).expand(-1, data.size(1))
    result = data.new_full(result_shape, 0)  # init empty result tensor
    count = data.new_full(result_shape, 0)
    result.scatter_add_(0, segment_ids, data)
    count.scatter_add_(0, segment_ids, torch.ones_like(data))
    return result / count.clamp(min=1)


class ListwiseModel(AutoModelForEmbedding):
    """
    segment_id
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def forward(
        self,
        inputs,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        return



from src.retrievals.losses.arcface import ArcFaceAdaptiveMarginLoss
from src.retrievals.losses.cosine_similarity import CosineSimilarity
from src.retrievals.losses.infonce import InfoNCE
from src.retrievals.losses.simcse import SimCSE
from src.retrievals.losses.triplet import TripletLoss



from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, TypeVar, Union

from transformers import AutoModel


class RAG(object):
    def __init__(self):
        pass

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: Union[str, Path],
        n_gpu: int = -1,
        verbose: int = 1,
        index_root: Optional[str] = None,
    ):
        instance = cls()
        instance.model = AutoModel()
        return instance

    @classmethod
    def from_index(cls, index_path: Union[str, Path], n_gpu: int = -1, verbose: int = 1):
        instance = cls()
        index_path = Path(index_path)
        instance.model = AutoModel()

        return instance

    def add_to_index(self):
        return

    def encode(self):
        return

    def index(self):
        return

    def search(self):
        return


class Generator(object):
    def __init__(self):
        pass



from src.retrievals.trainer.adversarial import AWP, EMA, FGM, PGD
from src.retrievals.trainer.custom_trainer import CustomTrainer, train_fn
from src.retrievals.trainer.trainer import RerankTrainer, RetrievalTrainer



from src.retrievals.data.collator import PairCollator, RerankCollator, TripletCollator
from src.retrievals.data.dataset import RerankDataset, RetrievalDataset
from src.retrievals.models.embedding_auto import AutoModelForEmbedding, PairwiseModel
from src.retrievals.models.match_auto import AutoModelForMatch
from src.retrievals.models.pooling import AutoPooling
from src.retrievals.models.rerank import RerankModel
from src.retrievals.trainer.custom_trainer import CustomTrainer
from src.retrievals.trainer.trainer import RerankTrainer, RetrievalTrainer



import os
import random
from dataclasses import dataclass
from typing import Iterable, List, Tuple

import datasets
from torch.utils.data import Dataset
from transformers import DataCollatorWithPadding, PreTrainedTokenizer


class RetrievalDataset(Dataset):
    def __init__(self, args):
        self.args = args
        if os.path.isdir(args.data_dir):
            train_datasets = []
            for file in os.listdir(args.data_dir):
                temp_dataset = datasets.load_dataset(
                    "json",
                    data_files=os.path.join(args.data_dir, file),
                    split="train",
                    cache_dir=args.cache_dir_data,
                )

                train_datasets.append(temp_dataset)
            self.dataset = datasets.concatenate_datasets(train_datasets)
        else:
            self.dataset = datasets.load_dataset("json", data_files=args.train_data, split="train")

        # self.tokenizer = tokenizer

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item):
        query = self.dataset[item]["query"]
        if isinstance(self.dataset[item]["pos"], Iterable):
            pos = random.choice(self.dataset[item]["pos"])
        else:
            pos = self.dataset[item]["pos"]
        if isinstance(self.dataset[item]["neg"], Iterable):
            neg = random.choice(self.dataset[item]["neg"])
        else:
            neg = self.dataset[item]["neg"][0]
        sample = {"query": query, "pos": pos, "neg": neg}
        return sample


class RerankDataset(Dataset):
    def __init__(self, args):
        self.args = args
        if os.path.isdir(args.data_dir):
            train_datasets = []
            for file in os.listdir(args.data_dir):
                temp_dataset = datasets.load_dataset(
                    "json",
                    data_files=os.path.join(args.data_dir, file),
                    split="train",
                    cache_dir=args.cache_dir_data,
                )

                train_datasets.append(temp_dataset)
            self.dataset = datasets.concatenate_datasets(train_datasets)
        else:
            self.dataset = datasets.load_dataset("json", data_files=args.train_data, split="train")

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item):
        query = self.dataset[item]["query"]
        passage = self.dataset[item]['passage']
        labels = self.dataset[item]['labels']
        sample = {"query": query, "passage": passage, "neg": labels}
        return sample



import gc
import logging
import math
import time
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import nn
from tqdm import tqdm

logger = logging.getLogger(__name__)


def train_fn(
    epoch,
    model,
    train_loader,
    optimizer,
    criterion=None,
    apex: bool = False,
    gradient_accumulation_steps: int = 1,
    max_grad_norm: float = 1,
    data_collator=None,
    batch_scheduler=None,
    fgm=None,
    awp=None,
    ema_inst=None,
    print_freq: int = 100,
    wandb: bool = False,
    device: str = "cuda",
    **kwargs,
):
    model.train()
    scaler = torch.cuda.amp.GradScaler(enabled=apex)
    losses = AverageMeter()
    start = time.time()
    global_step = 0
    save_step = int(len(train_loader) / 1)

    # loss_list = []
    # metrics_list = []

    for step, (inputs, labels) in enumerate(train_loader):
        if data_collator:
            inputs = data_collator(inputs)

        if isinstance(inputs, list):
            for input in inputs:
                for k, v in input.items():
                    input[k] = v.to(device)
        else:
            for k, v in inputs.items():
                inputs[k] = v.to(device)
        labels = labels.to(device)

        batch_size = labels.size(0)
        with torch.cuda.amp.autocast(enabled=apex):
            if criterion is None:
                preds = model(inputs, labels)
            else:
                preds = model(inputs)
            if isinstance(preds, dict) and "loss" in preds:
                loss = preds["loss"]
            else:
                loss = criterion(preds[0], preds[1])

        if gradient_accumulation_steps > 1:
            loss = loss / gradient_accumulation_steps
        losses.update(loss.item(), batch_size)
        scaler.scale(loss).backward()

        # if isinstance(preds, dict):
        #     acc = (preds["feature"].softmax(dim=1) * (labels > 0)).sum(axis=1).mean()
        # else:
        #     acc = (preds.softmax(dim=1) * (labels > 0)).sum(axis=1).mean()
        # acc = acc.detach().cpu().item()
        # acc_list.append(acc)

        if fgm:
            fgm.attack(epsilon=1.0)  # embedding被修改
            with torch.cuda.amp.autocast(enabled=apex):
                preds = model(inputs)
                if isinstance(preds, dict) and "loss" in preds:
                    loss_avg = preds["loss"]
                else:
                    loss_avg = criterion(preds, labels)
            if gradient_accumulation_steps > 1:
                loss_avg = loss_avg / gradient_accumulation_steps
            losses.update(loss_avg.item(), batch_size)
            scaler.scale(loss_avg).backward()
            fgm.restore()  # 恢复Embedding参数
        # ---------------------fgm-------------

        grad_norm = torch.nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
        if (step + 1) % gradient_accumulation_steps == 0:
            scaler.step(optimizer)
            scaler.update()

            if ema_inst:
                ema_inst.update()

            optimizer.zero_grad()
            global_step += 1
            if batch_scheduler:
                batch_scheduler.step()

        if step % print_freq == 0 or step == (len(train_loader) - 1):
            print(
                "Epoch: [{0}][{1}/{2}] "
                "Elapsed {remain:s} "
                "Loss: {loss.val:.4f}({loss.avg:.4f}) "
                "Grad: {grad_norm:.4f}  "
                "LR: {lr:.8f}  "
                # "ACC: {acc: .4f}"
                .format(
                    epoch + 1,
                    step,
                    len(train_loader),
                    remain=timeSince(start, float(step + 1) / len(train_loader)),
                    loss=losses,
                    grad_norm=grad_norm,
                    lr=batch_scheduler.get_lr()[0] if batch_scheduler else optimizer.param_groups[0]["lr"],
                    # acc=np.mean(acc_list),
                )
            )

        if wandb and step % print_freq == 0:
            print({"[loss": losses.val, "[lr": optimizer.param_groups[0]["lr"]})

        if (step + 1) % save_step == 0 and epoch > -1:
            if ema_inst:
                ema_inst.apply_shadow()

            if ema_inst:
                ema_inst.restore()

    return losses.avg


@torch.no_grad()
def valid_fn(
    valid_loader,
    model,
    criterion,
    eval_metrics=None,
    apex: bool = False,
    device="cuda",
    data_collator=None,
):
    model.eval()
    losses = AverageMeter()
    pred_array = torch.tensor([], device=device)
    label_array = np.array([])

    for step, (inputs, labels) in enumerate(valid_loader):
        if data_collator:
            inputs = data_collator(inputs)
        for k, v in inputs.items():
            inputs[k] = v.to(device)
        labels = labels.to(device)

        batch_size = labels.size(0)
        with torch.cuda.amp.autocast(enabled=apex):
            try:
                preds = model(inputs, labels=labels)
            except ValueError:
                preds = model(inputs)
            else:
                print('inputs of model should be with or without labels')

            if isinstance(preds, dict) and "loss" in preds:
                loss = preds["loss"]
                pred = preds["output"]
            else:
                loss = criterion(preds, labels)
                pred = preds

        losses.update(loss.item(), batch_size)
        pred_array = torch.cat((pred_array, pred), dim=0)
        label_array = np.append(label_array, np.array(labels))

    if eval_metrics is not None:
        score = eval_metrics(label_array, pred_array)
        return losses.avg, score

    return losses.avg


def inference_fn(test_loader, model, device):
    preds = []
    model.eval()
    model.to(device)
    tk0 = tqdm(test_loader, total=len(test_loader))
    for inputs in tk0:
        for k, v in inputs.items():
            inputs[k] = v.to(device)
        with torch.no_grad():
            y_preds = model(inputs)
        preds.append(y_preds.sigmoid().squeeze().to("cpu").numpy().reshape(-1))
    predictions = np.concatenate(preds)
    return predictions


class CustomTrainer(object):
    def __init__(self, model: Union[str, nn.Module], device=None, apex=False, teacher=None):
        if not device:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device
        self.model = model
        self.teacher = teacher
        self.apex = apex
        self.train_step = train_fn
        self.valid_step = valid_fn

    def train(
        self,
        train_loader,
        optimizer,
        epochs=10,
        criterion=None,
        scheduler=None,
        valid_loader=None,
        eval_metric=None,
        data_collator=None,
        max_grad_norm=10,
        **kwargs,
    ):
        # best_score = 0
        # fgm = FGM(model)
        # awp = None
        # ema_inst = EMA(model, 0.999)
        # ema_inst.register()

        for epoch in range(epochs):
            if "dynamic_margin" in kwargs.keys():
                if criterion:
                    criterion.set_margin(0.1 + epoch * 0.02)
                else:
                    self.model.loss_fn.set_margin(0.1 + epoch * 0.02)
                    logger.info(f"Epoch: {epoch}, Margin: {0.1 + epoch * 0.02}")

            # lr = scheduler(optimizer, epoch)

            # 放在这里, 是因为前面的set_margin也会产生新的参数位于cpu
            # 如果把参数直接放在cuda, 则导致parameter里没有出现该参数，无法分别设置学习率
            self.model = self.model.to(self.device)
            optimizer.zero_grad()
            self.train_step(
                epoch=epoch,
                train_loader=train_loader,
                model=self.model,
                criterion=criterion,
                optimizer=optimizer,
                batch_scheduler=scheduler,
                apex=self.apex,
                gradient_accumulation_steps=1,
                max_grad_norm=max_grad_norm,
                valid_loader=valid_loader,
                data_collator=data_collator,
                device=self.device,
            )

            if valid_loader is not None:
                self.valid_step(
                    valid_loader,
                    self.model,
                    criterion=criterion,
                    apex=self.apex,
                    device=self.device,
                    data_collator=None,
                )

        torch.cuda.empty_cache()
        gc.collect()

    def predict(self, test_loader):
        self.model.eval()
        tbar = tqdm(test_loader)

        vectors = []
        with torch.no_grad():
            for idx, (inputs, labels) in enumerate(tbar):
                for k, v in inputs.items():
                    inputs[k] = v.to(self.device)

                vec = self.model(inputs)
                if isinstance(vec, dict):
                    vec = vec["embeddings"]
                vectors.append(vec.detach().cpu().numpy())

        V = np.concatenate(vectors)
        return V

    def save_state(self, output_dir: Optional[str] = None):
        # TODO: remove the loss_fn
        if not output_dir:
            output_dir = "./model.pth"
        torch.save(self.model.state_dict(), output_dir)

    def save_model(self, output_dir: Optional[str] = None, state_dict=None):
        pass


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def asMinutes(s):
    m = math.floor(s / 60)
    s -= m * 60
    return "%dm %ds" % (m, s)


def timeSince(since, percent):
    now = time.time()
    s = now - since
    es = s / (percent)
    rs = es - s
    return "%s (remain %s)" % (asMinutes(s), asMinutes(rs))



import logging
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import torch
import torch.nn as nn
from transformers import AutoConfig, AutoModel, AutoTokenizer

from src.retrievals.data.collator import RerankCollator
from src.retrievals.models.embedding_auto import get_device_name
from src.retrievals.models.pooling import AutoPooling

logger = logging.getLogger(__name__)


class RerankModel(nn.Module):
    def __init__(
        self,
        model_name_or_path: str = None,
        pooling_method: str = 'mean',
        loss_fn: Union[nn.Module, Callable] = None,
        max_length: Optional[int] = None,
        use_fp16: bool = False,
        use_lora: bool = False,
        lora_config=None,
        gradient_checkpointing: bool = False,
        device: Optional[str] = None,
        trust_remote_code: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, return_tensors=False, trust_remote_code=trust_remote_code
        )

        self.model = AutoModel.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code)
        if gradient_checkpointing:
            self.model.graident_checkpointing_enable()
        if device is None:
            self.device = get_device_name()
        else:
            self.device = device

        if use_fp16:
            self.model.half()
        if use_lora:
            # peft config and wrapping
            from peft import LoraConfig, TaskType, get_peft_model

            if not lora_config:
                raise ValueError("If use_lora is true, please provide a valid lora_config")
            self.model = get_peft_model(self.model, lora_config)
            self.model.print_trainable_parameters()

        self.pooling = AutoPooling(pooling_method)
        num_features = self.backbone.config.hidden_size
        self.classifier = nn.Linear(num_features, 1)
        self._init_weights(self.classifier)
        self.loss_fn = loss_fn

        if max_length is None:
            if (
                hasattr(self.model, "config")
                and hasattr(self.model.config, "max_position_embeddings")
                and hasattr(self.tokenizer, "model_max_length")
            ):
                max_length = min(self.model.config.max_position_embeddings, self.tokenizer.model_max_length)

        self.max_length = max_length

    def _init_weights(self, module: nn.Module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def encode(self, input_ids: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        outputs = self.model(input_ids, attention_mask, output_hidden_states=False)
        encoder_layer = outputs.last_hidden_state
        embeddings = self.pooling(encoder_layer, attention_mask)
        return embeddings

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        labels: Optional[torch.Tensor] = None,
        return_dict: Optional[bool] = False,
        **kwargs,
    ) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
        features = self.encode(input_ids=input_ids, attention_mask=attention_mask)
        logits = self.classifier(features).reshape(-1)

        if labels is not None:
            if not self.loss_fn:
                logger.warning('loss_fn is not setup, use BCEWithLogitsLoss')
                self.loss_fn = nn.BCEWithLogitsLoss(reduction='mean')

        loss = self.loss_fn(logits, labels)
        if return_dict:
            outputs_dict = dict()
            outputs_dict['loss'] = loss
            outputs_dict['logits'] = logits
            return outputs_dict
        return loss

    def compute_score(
        self,
        text: Union[List[str], str],
        text_pair: Union[List[str], str],
        data_collator: Optional[RerankCollator] = None,
        batch_size: int = 128,
        show_progress_bar: bool = None,
        **kwargs,
    ):
        if isinstance(text, str):
            text = [text]
        if isinstance(text_pair, str):
            text_pair = [text_pair]
        assert len(text) == len(text_pair), f"Length of text {len(text)} and text_pair {len(text_pair)} should be same"

        if not data_collator:
            data_collator = RerankCollator(tokenizer=self.tokenizer)

        with torch.no_grad():
            scores_list: List = []
            for i in range(0, len(text), batch_size):
                text_batch = [{'query': text[i], 'passage': text_pair[i]} for i in range(i, i + batch_size)]
                batch = data_collator(text_batch)
                scores = self.model(batch, return_dict=True).logits.view(-1).float()
                scores = torch.sigmoid(scores)
                scores_list.extend(scores.cpu().numpy().tolist())

        return scores_list

    def rerank(
        self,
        query: Union[List[str], str],
        passages: List[str],
        data_collator: Optional[RerankCollator] = None,
        batch_size: int = 32,
        show_progress_bar: bool = None,
        **kwargs,
    ):

        merge_scores = self.compute_score(query, passages, data_collator, batch_size, show_progress_bar)

        merge_scores_argsort = np.argsort(merge_scores)[::-1]
        sorted_passages = []
        sorted_scores = []
        for mid in merge_scores_argsort:
            sorted_scores.append(merge_scores[mid])
            sorted_passages.append(passages[mid])

        return {
            'rerank_passages': sorted_passages,
            'rerank_scores': sorted_scores,
            'rerank_ids': merge_scores_argsort.tolist(),
        }

    def save(self, path: str):
        """
        Saves all model and tokenizer to path
        """
        if path is None:
            return

        logger.info("Save model to {}".format(path))
        self.model.save_pretrained(path)
        self.tokenizer.save_pretrained(path)

    def save_pretrained(self, path: str):
        """
        Same function to save
        """
        return self.save(path)
